from typing import Sequence
import numpy as np


class YoungLaplaceShape:
    def __init__(self, bond: float) -> None: ...
    def __call__(self, s: float) -> np.ndarray: ...
    def DBo(self, s: float) -> np.ndarray: ...
    def z_inv(self, s: float) -> float: ...
    def closest(self, v: Sequence[float]) -> float: ...
    def volume(self, s: float) -> float: ...
    def surface_area(self, s: float) -> float: ...
    @property
    def bond(self) -> float: ...
